Attribute VB_Name = "modFFT"
'--------------------------------------------------------------------
' VB FFT Release 2-B
' by Murphy McCauley (MurphyMc@Concentric.NET)
' 10/01/99
'--------------------------------------------------------------------
' About:
' This code is very, very heavily based on Don Cross's fourier.pas
' Turbo Pascal Unit for calculating the Fast Fourier Transform.
' I've not implemented all of his functions, though I may well do
' so in the future.
' For more info, you can contact me by email, check my website at:
' http://www.fullspectrum.com/deeth/
' or check Don Cross's FFT web page at:
' http://www.intersrv.com/~dcross/fft.html
' You also may be intrested in the FFT.DLL that I put together based
' on Don Cross's FFT C code.  It's callable with Visual Basic and
' includes VB declares.  You can get it from either website.
'--------------------------------------------------------------------
' History of Release 2-B:
' Fixed a couple of errors that resulted from me mucking about with
'   variable names after implementation and not re-checking.  BAD ME.
'  --------
' History of Release 2:
' Added FrequencyOfIndex() which is Don Cross's Index_to_frequency().
' FourierTransform() can now do inverse transforms.
' Added CalcFrequency() which can do a transform for a single
'   frequency.
'--------------------------------------------------------------------
' Usage:
' The useful functions are:
' FourierTransform() performs a Fast Fourier Transform on an pair of
'  Double arrays -- one real, one imaginary.  Don't want/need
'  imaginary numbers?  Just use an array of 0s.  This function can
'  also do inverse FFTs.
' FrequencyOfIndex() can tell you what actual frequency a given index
'  corresponds to.
' CalcFrequency() transforms a single frequency.
'--------------------------------------------------------------------
' Notes:
' All arrays must be 0 based (i.e. Dim TheArray(0 To 1023) or
'  Dim TheArray(1023)).
' The number of samples must be a power of two (i.e. 2^x).
' FrequencyOfIndex() and CalcFrequency() haven't been tested much.
' Use this ENTIRELY AT YOUR OWN RISK.
'--------------------------------------------------------------------

Option Explicit

Private Const PI        As Double = 3.14159265358979

Private m_lngPowers(16) As Long

Private Function NumberOfBitsNeeded(ByVal PowerOfTwo As Long) As Long
    Dim i               As Long
    
    If m_lngPowers(0) = 0 Then
        For i = 0 To UBound(m_lngPowers)
            m_lngPowers(i) = 2 ^ i
        Next
    End If

    For i = 0 To 16
        If (PowerOfTwo And m_lngPowers(i)) <> 0 Then
            NumberOfBitsNeeded = i
            Exit Function
        End If
    Next
End Function

Private Function IsPowerOfTwo(ByVal X As Long) As Boolean
    If (X < 2) Then Exit Function
    IsPowerOfTwo = Not (X And (X - 1))
End Function

Private Function ReverseBits(ByVal Index As Long, ByVal NumBits As Long) As Long
    Dim i As Long, Rev As Long

    For i = 0 To NumBits - 1
        Rev = (Rev * 2) Or (Index And 1)
        Index = Index \ 2
    Next

    ReverseBits = Rev
End Function

Public Sub FourierTransform( _
    ByVal NumSamples As Long, _
    RealIn() As Integer, _
    ImageIn() As Double, _
    RealOut() As Double, _
    ImagOut() As Double, _
    Optional InverseTransform As Boolean = False _
)

    Dim i           As Long, j          As Long
    Dim K           As Long, n          As Long

    Dim BlockSize   As Long, BlockEnd   As Long

    Dim DeltaAngle  As Double, DeltaAr  As Double
    Dim Alpha       As Double, Beta     As Double

    Dim TR          As Double, TI       As Double
    Dim AR          As Double, AI       As Double
    
    Dim AngleNumerator                  As Double

    Dim NumBits                         As Long

    If InverseTransform Then
        AngleNumerator = -2# * PI
    Else
        AngleNumerator = 2# * PI
    End If

'    If (Not IsPowerOfTwo(NumSamples)) Or (NumSamples < 2) Then
'        Exit Sub
'    End If

    NumBits = NumberOfBitsNeeded(NumSamples)

    For i = 0 To (NumSamples - 1)
        j = ReverseBits(i, NumBits)
        RealOut(j) = RealIn(i)
        ImagOut(j) = ImageIn(i)
    Next

    BlockEnd = 1
    BlockSize = 2

    Do While BlockSize <= NumSamples
        DeltaAngle = AngleNumerator / BlockSize
        Alpha = Sin(0.5 * DeltaAngle)
        Alpha = 2# * Alpha * Alpha
        Beta = Sin(DeltaAngle)

        i = 0
        Do While i < NumSamples
            AR = 1#
            AI = 0#
            
            j = i
            For n = 0 To BlockEnd - 1
                K = j + BlockEnd
                TR = AR * RealOut(K) - AI * ImagOut(K)
                TI = AI * RealOut(K) + AR * ImagOut(K)
                RealOut(K) = RealOut(j) - TR
                ImagOut(K) = ImagOut(j) - TI
                RealOut(j) = RealOut(j) + TR
                ImagOut(j) = ImagOut(j) + TI
                DeltaAr = Alpha * AR + Beta * AI
                AI = AI - (Alpha * AI - Beta * AR)
                AR = AR - DeltaAr
                j = j + 1
            Next

            i = i + BlockSize
        Loop

        BlockEnd = BlockSize
        BlockSize = BlockSize * 2
    Loop

    If InverseTransform Then
        For i = 0 To NumSamples - 1
            RealOut(i) = RealOut(i) / NumSamples
            ImagOut(i) = ImagOut(i) / NumSamples
        Next
    End If
End Sub

Public Function Hanning(ByVal X As Long, ByVal Length As Long) As Double
    Hanning = 0.5 * (1 - Cos((2 * PI * X) / Length))
End Function
